// Copyright 2000-2021 JetBrains s.r.o. and contributors. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.openapi.module.impl;

import com.intellij.ProjectTopics;
import com.intellij.configurationStore.StateStorageManagerKt;
import com.intellij.openapi.Disposable;
import com.intellij.openapi.application.Application;
import com.intellij.openapi.application.ApplicationInfo;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.application.ReadAction;
import com.intellij.openapi.components.ComponentManager;
import com.intellij.openapi.components.PersistentStateComponent;
import com.intellij.openapi.components.ProjectComponent;
import com.intellij.openapi.components.impl.stores.IComponentStore;
import com.intellij.openapi.components.impl.stores.ModuleStore;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.module.Module;
import com.intellij.openapi.module.*;
import com.intellij.openapi.progress.*;
import com.intellij.openapi.progress.util.ProgressWrapper;
import com.intellij.openapi.project.ModuleListener;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.roots.ModuleRootManager;
import com.intellij.openapi.roots.ex.ProjectRootManagerEx;
import com.intellij.openapi.roots.impl.ModifiableModelCommitter;
import com.intellij.openapi.roots.impl.ModuleRootManagerImpl;
import com.intellij.openapi.util.*;
import com.intellij.openapi.util.io.FileUtil;
import com.intellij.openapi.util.io.FileUtilRt;
import com.intellij.openapi.util.registry.Registry;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.openapi.vfs.StandardFileSystems;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.openapi.vfs.VirtualFileManager;
import com.intellij.openapi.vfs.pointers.VirtualFilePointer;
import com.intellij.openapi.vfs.pointers.VirtualFilePointerManager;
import com.intellij.projectModel.ProjectModelBundle;
import com.intellij.util.ConcurrencyUtil;
import com.intellij.util.ObjectUtils;
import com.intellij.util.SmartList;
import com.intellij.util.concurrency.AppExecutorUtil;
import com.intellij.util.containers.ContainerUtil;
import com.intellij.util.containers.Interner;
import com.intellij.util.graph.*;
import it.unimi.dsi.fastutil.Hash;
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenCustomHashMap;
import it.unimi.dsi.fastutil.objects.ObjectArrays;
import org.jdom.Element;
import org.jetbrains.annotations.ApiStatus;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.jetbrains.annotations.SystemIndependent;
import org.jetbrains.jps.model.serialization.JpsProjectLoader;

import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.NoSuchFileException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.stream.Collectors;

import static com.intellij.openapi.module.impl.ExternalModuleListStorageKt.getFilteredModuleList;

/**
 * This class isn't used in the new implementation of project model, which is based on {@link com.intellij.workspaceModel.ide Workspace Model}.
 * It shouldn't be used directly, its base class {@link ModuleManagerEx} should be used instead.
 */
@ApiStatus.Internal
public abstract class ModuleManagerImpl extends ModuleManagerEx implements Disposable, PersistentStateComponent<Element>, ProjectComponent {
  private static final Logger LOG = Logger.getInstance(ModuleManagerImpl.class);
  private static final Key<String> DISPOSED_MODULE_NAME = Key.create("DisposedNeverAddedModuleName");

  protected final Project myProject;
  private final ProjectRootManagerEx myProjectRootManager;
  protected volatile ModuleModelImpl myModuleModel = new ModuleModelImpl(this);

  private Set<ModulePath> myModulePathsToLoad;
  private final Set<ModulePath> myFailedModulePaths = new HashSet<>();
  private final Map<String, UnloadedModuleDescriptionImpl> myUnloadedModules = new LinkedHashMap<>();
  private boolean myModulesLoaded;

  public static ModuleManagerImpl getInstanceImpl(Project project) {
    return (ModuleManagerImpl)getInstance(project);
  }

  public ModuleManagerImpl(@NotNull Project project) {
    myProject = project;
    myProjectRootManager = ProjectRootManagerEx.getInstanceEx(project);
  }

  @Override
  public void projectOpened() {
    fireModulesAdded();

    for (Module module : myModuleModel.getModules()) {
      ((ModuleEx)module).projectOpened();
    }
  }

  @Override
  public void projectClosed() {
    for (Module module : myModuleModel.getModules()) {
      ((ModuleEx)module).projectClosed();
    }
  }

  protected void cleanCachedStuff() {
    myCachedModuleComparator = null;
    myCachedSortedModules = null;
    myCachedModuleProductionGraph = null;
    myCachedModuleTestGraph = null;
  }

  @Override
  public void dispose() {
    myModuleModel.disposeModel();
    myProjectRootManager.assertListenersAreDisposed();
  }

  @Override
  public Element getState() {
    final Element e = new Element("state");
    writeExternal(e, getFilteredModuleList(myProject, getModules(), false));
    return e;
  }

  private static final class ModuleGroupInterner {
    private final Interner<String> groups = Interner.createStringInterner();
    @SuppressWarnings("unchecked")
    private final Map<String[], String[]> paths = new Object2ObjectOpenCustomHashMap<>((Hash.Strategy<? super String[]>)ObjectArrays.HASH_STRATEGY);

    private void setModuleGroupPath(@NotNull ModifiableModuleModel model, @NotNull Module module, String @Nullable [] group) {
      String[] cached = group == null ? null : paths.get(group);
      if (cached == null && group != null) {
        cached = new String[group.length];
        for (int i = 0; i < group.length; i++) {
          String g = group[i];
          cached[i] = groups.intern(g);
        }
        paths.put(cached, cached);
      }
      model.setModuleGroupPath(module, cached);
    }
  }

  @Override
  public void loadState(@NotNull Element state) {
    Set<ModulePath> files = getPathsToModuleFiles(state);
    Set<ModulePath> externalModules = myProject.getComponent(ExternalModuleListStorage.class).getExternalModules();
    if (externalModules != null) {
      files.addAll(externalModules);
    }
    loadState(files);
  }

  @Override
  public void noStateLoaded() {
    // if there are only external modules, loadState will be not called
    Set<ModulePath> externalModules = myProject.getComponent(ExternalModuleListStorage.class).getExternalModules();
    if (externalModules != null) {
      loadState(new LinkedHashSet<>(externalModules));
    }
  }

  private void loadState(@NotNull Set<ModulePath> modulePaths) {
    boolean isFirstLoadState = myModulePathsToLoad == null;
    myModulePathsToLoad = modulePaths;
    Set<String> unloadedModuleNames = UnloadedModulesListStorage.getInstance(myProject).getUnloadedModuleNames();
    Iterator<ModulePath> iterator = myModulePathsToLoad.iterator();
    List<ModulePath> unloadedModulePaths = new ArrayList<>();
    while (iterator.hasNext()) {
      ModulePath modulePath = iterator.next();
      if (unloadedModuleNames.contains(modulePath.getModuleName())) {
        unloadedModulePaths.add(modulePath);
        iterator.remove();
      }
    }
    List<UnloadedModuleDescriptionImpl> unloaded = new ArrayList<>(UnloadedModuleDescriptionImpl.createFromPaths(unloadedModulePaths, this));
    if (!unloaded.isEmpty()) {
      unloadNewlyAddedModulesIfPossible(myModulePathsToLoad, unloaded);
    }
    myUnloadedModules.clear();
    for (UnloadedModuleDescriptionImpl description : unloaded) {
      myUnloadedModules.put(description.getName(), description);
    }

    if (isFirstLoadState) {
      // someone else must call loadModules in a appropriate time (e.g. on projectComponentsInitialized)
      return;
    }

    final ModifiableModuleModel model = getModifiableModel();
    Module[] existingModules = model.getModules();
    ModuleGroupInterner groupInterner = new ModuleGroupInterner();

    Map<String, ModulePath> modulePathMap = new HashMap<>(myModulePathsToLoad.size());
    for (ModulePath modulePath : myModulePathsToLoad) {
      modulePathMap.put(modulePath.getPath(), modulePath);
    }

    for (Module existingModule : existingModules) {
      ModulePath correspondingPath = modulePathMap.remove(existingModule.getModuleFilePath());
      if (correspondingPath == null) {
        model.disposeModule(existingModule);
      }
      else {
        myModulePathsToLoad.remove(correspondingPath);

        String groupStr = correspondingPath.getGroup();
        String[] group = groupStr == null ? null : groupStr.split(MODULE_GROUP_SEPARATOR);
        if (!Arrays.equals(group, model.getModuleGroupPath(existingModule))) {
          groupInterner.setModuleGroupPath(model, existingModule, group);
        }
      }
    }

    loadModules((ModuleModelImpl)model);
    ApplicationManager.getApplication().runWriteAction(model::commit);
    // clear only if successfully loaded
    myModulePathsToLoad.clear();
  }

  @NotNull
  // returns mutable linked hash set
  public static Set<ModulePath> getPathsToModuleFiles(@NotNull Element element) {
    Set<ModulePath> paths = new LinkedHashSet<>();
    final Element modules = element.getChild(JpsProjectLoader.MODULES_TAG);
    if (modules != null) {
      for (final Element moduleElement : modules.getChildren(JpsProjectLoader.MODULE_TAG)) {
        final String fileUrlValue = moduleElement.getAttributeValue(JpsProjectLoader.FILE_URL_ATTRIBUTE);
        final String filepath;
        if (fileUrlValue == null) {
          // support for older formats
          filepath = moduleElement.getAttributeValue(JpsProjectLoader.FILE_PATH_ATTRIBUTE);
        }
        else {
          filepath = VirtualFileManager.extractPath(fileUrlValue);
        }
        paths.add(new ModulePath(FileUtilRt.toSystemIndependentName(Objects.requireNonNull(filepath)), moduleElement.getAttributeValue(JpsProjectLoader.GROUP_ATTRIBUTE)));
      }
    }
    return paths;
  }

  protected void loadModules(@NotNull ModuleModelImpl moduleModel) {
    myFailedModulePaths.clear();

    if (myModulePathsToLoad == null || myModulePathsToLoad.isEmpty()) {
      myModulesLoaded = true;
      return;
    }

    myFailedModulePaths.addAll(myModulePathsToLoad);

    ProgressIndicator globalIndicator = ProgressIndicatorProvider.getGlobalProgressIndicator();
    ProgressIndicator progressIndicator = myProject.isDefault() || globalIndicator == null ? new EmptyProgressIndicator() : globalIndicator;
    progressIndicator.setText(ProjectModelBundle.message("progress.text.loading.modules"));
    progressIndicator.setText2("");

    List<ModuleLoadingErrorDescription> errors = Collections.synchronizedList(new ArrayList<>());

    double progressStep = (1 - progressIndicator.getFraction()) / myModulePathsToLoad.size();

    boolean isParallel = Registry.is("parallel.modules.loading") && !ApplicationManager.getApplication().isDispatchThread();
    ExecutorService service = isParallel ? AppExecutorUtil.createBoundedApplicationPoolExecutor("ModuleManager Loader", Math.min(2, Runtime.getRuntime().availableProcessors()))
                                         : ConcurrencyUtil.newSameThreadExecutorService();
    List<Pair<Future<Module>, ModulePath>> tasks = new ArrayList<>();
    Set<String> paths = new HashSet<>(myModulePathsToLoad.size());
    for (ModulePath modulePath : myModulePathsToLoad) {
      if (progressIndicator.isCanceled()) {
        break;
      }

      String path = modulePath.getPath();
      if (!paths.add(path)) {
        continue;
      }

      tasks.add(new Pair<>(service.submit(() -> {
        progressIndicator.setFraction(progressIndicator.getFraction() + progressStep);
        return ProgressManager.getInstance().runProcess(() -> {
          try {
            return myProject.isDisposed() ? null : loadModuleInternal(Paths.get(path), this);
          }
          catch (IOException e) {
            reportError(errors, modulePath, e);
          }
          catch (ProcessCanceledException ignore) {
          }
          catch (Exception e) {
            LOG.error(e);
          }
          return null;
        }, ProgressWrapper.wrap(progressIndicator));
      }), modulePath));
    }

    ModuleGroupInterner groupInterner = new ModuleGroupInterner();
    List<Module> modulesWithUnknownTypes = new SmartList<>();
    for (Pair<Future<Module>, ModulePath> task : tasks) {
      if (progressIndicator.isCanceled()) {
        break;
      }

      try {
        Module module = task.first.get();
        if (module == null) {
          continue;
        }

        moduleModel.addModule(module);

        if (isUnknownModuleType(module)) {
          modulesWithUnknownTypes.add(module);
        }
        ModulePath modulePath = task.second;
        String groupPathString = modulePath.getGroup();
        if (groupPathString != null) {
          // model should be updated too
          groupInterner.setModuleGroupPath(moduleModel, module, groupPathString.split(MODULE_GROUP_SEPARATOR));
        }
        myFailedModulePaths.remove(modulePath);
      }
      catch (Exception e) {
        LOG.error(e);
      }
    }
    service.shutdown();

    progressIndicator.checkCanceled();

    myModulesLoaded = true;

    Application app = ApplicationManager.getApplication();
    if (app.isInternal() || app.isEAP() || ApplicationInfo.getInstance().getBuild().isSnapshot()) {
      Module[] modules = moduleModel.getModules();
      Map<String, Module> track = new HashMap<>(modules.length);
      for (Module module : modules) {
        for (String url : ModuleRootManager.getInstance(module).getContentRootUrls()) {
          Module oldModule = track.put(url, module);
          if (oldModule != null) {
            LOG.warn("Module '" + module.getName() + "' and module '" + oldModule.getName() + "' have the same content root: " + url);
          }
        }
      }
    }

    onModuleLoadErrors(moduleModel, errors);

    showUnknownModuleTypeNotification(modulesWithUnknownTypes);
  }

  private static @NotNull Module loadModuleInternal(@NotNull Path file, @NotNull ModuleManagerImpl manager) throws IOException {
    // we cannot call refreshAndFindFileByPath during module init under read action because it is forbidden
    String systemIndependentPath = file.toString().replace(File.separatorChar, '/');
    VirtualFile virtualFile = StandardFileSystems.local().refreshAndFindFileByPath(systemIndependentPath);
    if (virtualFile != null) {
      // otherwise virtualFile.contentsToByteArray() will query expensive FileTypeManager.getInstance()).getByFile()
      virtualFile.setCharset(StandardCharsets.UTF_8, null, false);
    }
    return ReadAction.compute(() -> {
      ModuleEx module = manager.createAndLoadModule(systemIndependentPath);
      initModule(module, () -> ((ModuleStore)module.getService(IComponentStore.class)).setPath(file, virtualFile, false));
      return module;
    });
  }

  private static void initModule(@NotNull ModuleEx module, @NotNull Runnable beforeComponentCreation) {
    try {
      module.init(beforeComponentCreation);
    }
    catch (Throwable e) {
      disposeModuleLater(module);
      throw e;
    }
  }

  private void reportError(@NotNull List<ModuleLoadingErrorDescription> errors, @NotNull ModulePath modulePath, @NotNull Exception e) {
    errors.add(new ModuleLoadingErrorDescription(ProjectModelBundle.message("module.cannot.load.error", modulePath.getPath(), e.getMessage()), modulePath, this));
  }

  public int getModulePathsCount() {
    return myModulePathsToLoad == null ? 0 : myModulePathsToLoad.size();
  }

  @Override
  public boolean areModulesLoaded() {
    return myModulesLoaded;
  }

  protected boolean isUnknownModuleType(@NotNull Module module) {
    return false;
  }

  protected void showUnknownModuleTypeNotification(@NotNull List<? extends Module> types) {
  }

  protected void fireModuleRemoved(@NotNull Module module) {
    myProject.getMessageBus().syncPublisher(ProjectTopics.MODULES).moduleRemoved(myProject, module);
  }

  protected void fireBeforeModuleRemoved(@NotNull Module module) {
    myProject.getMessageBus().syncPublisher(ProjectTopics.MODULES).beforeModuleRemoved(myProject, module);
  }

  protected void fireModulesRenamed(@NotNull List<Module> modules, @NotNull final Map<Module, String> oldNames) {
    if (modules.isEmpty()) {
      return;
    }

    try {
      for (Module module : getModules()) {
        ModuleRootManager moduleRootManager = ModuleRootManager.getInstance(module);
        if (moduleRootManager instanceof ModuleRootManagerImpl) {
          // platform in any case will check that iml is actually modified
          ((ModuleRootManagerImpl)moduleRootManager).stateChanged();
        }
      }
    }
    catch (Exception e) {
      LOG.error(e);
    }

    myProject.getMessageBus().syncPublisher(ProjectTopics.MODULES).modulesRenamed(myProject, modules, oldNames::get);
  }

  private void onModuleLoadErrors(@NotNull ModuleModelImpl moduleModel, @NotNull List<ModuleLoadingErrorDescription> errors) {
    if (errors.isEmpty()) return;

    moduleModel.myModulesCache = null;
    for (ModuleLoadingErrorDescription error : errors) {
      final Module module = moduleModel.getModuleByFilePath(error.getModulePath().getPath());
      if (module != null) {
        moduleModel.myModules.remove(module.getName());
        disposeModuleLater(module);
      }
    }

    fireModuleLoadErrors(errors);
  }

  private static void disposeModuleLater(@NotNull ComponentManager module) {
    ApplicationManager.getApplication().invokeLater(() -> Disposer.dispose(module), module.getDisposed());
  }

  // overridden in Upsource
  protected void fireModuleLoadErrors(@NotNull List<ModuleLoadingErrorDescription> errors) {
    if (ApplicationManager.getApplication().isHeadlessEnvironment() && !ApplicationManager.getApplication().isUnitTestMode()) {
      throw new RuntimeException(errors.get(0).getDescription());
    }

    ProjectLoadingErrorsNotifier.getInstance(myProject).registerErrors(errors);
  }

  void removeFailedModulePath(@NotNull ModulePath modulePath) {
    myFailedModulePaths.remove(modulePath);
    incModificationCount();
  }

  @Override
  @NotNull
  public ModifiableModuleModel getModifiableModel() {
    ApplicationManager.getApplication().assertReadAccessAllowed();
    return new ModuleModelImpl(myModuleModel);
  }

  private static class ModuleSaveItem extends SaveItem {
    private final Module myModule;
    private final ModuleManager myModuleManager;

    ModuleSaveItem(@NotNull Module module, @NotNull ModuleManager moduleManager) {
      myModule = module;
      myModuleManager = moduleManager;
    }

    @Override
    @NotNull
    protected String getModuleName() {
      return myModule.getName();
    }

    @Override
    protected String getGroupPathString() {
      String[] groupPath = myModuleManager.getModuleGroupPath(myModule);
      return groupPath != null ? StringUtil.join(groupPath, MODULE_GROUP_SEPARATOR) : null;
    }

    @Override
    @NotNull
    protected String getModuleFilePath() {
      return myModule.getModuleFilePath();
    }
  }

  public void writeExternal(@NotNull Element element, @NotNull List<? extends Module> collection) {
    writeExternal(element, collection, this);
  }

  @ApiStatus.Internal
  public static void writeExternal(@NotNull Element element,
                                   @NotNull List<? extends Module> collection,
                                   ModuleManagerEx moduleManager) {
    Collection<ModulePath> failedModulePaths = moduleManager.getFailedModulePaths();
    Collection<UnloadedModuleDescription> unloadedModuleDescriptions = moduleManager.getUnloadedModuleDescriptions();

    List<SaveItem> sorted = new ArrayList<>(collection.size() + failedModulePaths.size() + unloadedModuleDescriptions.size());
    for (Module module : collection) {
      sorted.add(new ModuleSaveItem(module, moduleManager));
    }
    for (ModulePath modulePath : failedModulePaths) {
      sorted.add(new ModulePathSaveItem(modulePath));
    }
    for (UnloadedModuleDescription description : unloadedModuleDescriptions) {
      sorted.add(new ModulePathSaveItem(((UnloadedModuleDescriptionImpl)description).getModulePath()));
    }

    if (!sorted.isEmpty()) {
      sorted.sort(Comparator.comparing(SaveItem::getModuleName));

      Element modules = new Element(JpsProjectLoader.MODULES_TAG);
      for (SaveItem saveItem : sorted) {
        saveItem.writeExternal(modules);
      }
      element.addContent(modules);
    }
  }

  @Override
  public @NotNull Module newModule(@NotNull String filePath, @NotNull String moduleTypeId) {
    incModificationCount();
    ModifiableModuleModel modifiableModel = getModifiableModel();
    Module module = modifiableModel.newModule(filePath, moduleTypeId);
    modifiableModel.commit();
    return module;
  }

  @Override
  public @NotNull Module newModule(@NotNull Path filePath, @NotNull String moduleTypeId) {
    incModificationCount();
    ModifiableModuleModel modifiableModel = getModifiableModel();
    Module module = modifiableModel.newModule(filePath, moduleTypeId);
    modifiableModel.commit();
    return module;
  }

  @Override
  @NotNull
  public Module newNonPersistentModule(@NotNull String moduleName, @NotNull String id) {
    incModificationCount();
    final ModifiableModuleModel modifiableModel = getModifiableModel();
    final Module module = modifiableModel.newNonPersistentModule(moduleName, id);
    modifiableModel.commit();
    return module;
  }

  @Override
  @NotNull
  public Module loadModule(@NotNull String filePath) throws IOException, ModuleWithNameAlreadyExists {
    incModificationCount();
    ModifiableModuleModel modifiableModel = getModifiableModel();
    Module module = modifiableModel.loadModule(filePath);
    modifiableModel.commit();
    return module;
  }

  @Override
  @NotNull
  public Module loadModule(@NotNull Path file) throws IOException, ModuleWithNameAlreadyExists {
    incModificationCount();
    ModifiableModuleModel modifiableModel = getModifiableModel();
    Module module = modifiableModel.loadModule(file);
    modifiableModel.commit();
    return module;
  }

  @Override
  public void disposeModule(@NotNull Module module) {
    ApplicationManager.getApplication().runWriteAction(() -> {
      final ModifiableModuleModel modifiableModel = getModifiableModel();
      modifiableModel.disposeModule(module);
      modifiableModel.commit();
    });
  }

  @Override
  public Module @NotNull [] getModules() {
    ModuleModelImpl model = myModuleModel;
    if (model.myIsWritable) {
      ApplicationManager.getApplication().assertReadAccessAllowed();
    }
    return model.getModules();
  }

  private volatile Module[] myCachedSortedModules;
  private volatile Graph<Module> myCachedModuleProductionGraph;
  private volatile Graph<Module> myCachedModuleTestGraph;

  @Override
  public Module @NotNull [] getSortedModules() {
    ApplicationManager.getApplication().assertReadAccessAllowed();
    deliverPendingEvents();
    Module[] sortedModules = myCachedSortedModules;
    if (sortedModules == null) {
      sortedModules = myModuleModel.getSortedModules();
      myCachedSortedModules = sortedModules;
    }
    return sortedModules;
  }

  @Override
  public Module findModuleByName(@NotNull String name) {
    return myModuleModel.findModuleByName(name);
  }

  private volatile Comparator<Module> myCachedModuleComparator;

  @Override
  @NotNull
  public Comparator<Module> moduleDependencyComparator() {
    ApplicationManager.getApplication().assertReadAccessAllowed();
    deliverPendingEvents();
    Comparator<Module> cachedModuleComparator = myCachedModuleComparator;
    if (cachedModuleComparator == null) {
      myCachedModuleComparator = cachedModuleComparator = myModuleModel.moduleDependencyComparator();
    }
    return cachedModuleComparator;
  }

  protected void deliverPendingEvents() {
  }

  @Override
  @NotNull
  public Graph<Module> moduleGraph() {
    return moduleGraph(true);
  }

  @NotNull
  @Override
  public Graph<Module> moduleGraph(boolean includeTests) {
    ApplicationManager.getApplication().assertReadAccessAllowed();

    Graph<Module> graph = includeTests ? myCachedModuleTestGraph : myCachedModuleProductionGraph;
    if (graph != null) return graph;

    graph = myModuleModel.moduleGraph(includeTests);
    if (includeTests) {
      myCachedModuleTestGraph = graph;
    }
    else {
      myCachedModuleProductionGraph = graph;
    }

    return graph;
  }

  @Override
  @NotNull
  public List<Module> getModuleDependentModules(@NotNull Module module) {
    List<Module> result = new SmartList<>();
    for (Module aModule : getModules()) {
      if (isModuleDependsOn(aModule, module)) {
        result.add(aModule);
      }
    }
    return result;
  }

  @Override
  public boolean isModuleDependent(@NotNull Module module, @NotNull Module onModule) {
    ApplicationManager.getApplication().assertReadAccessAllowed();
    return isModuleDependsOn(module, onModule);
  }

  private static boolean isModuleDependsOn(@NotNull Module module, @NotNull Module onModule) {
    return ModuleRootManager.getInstance(module).isDependsOn(onModule);
  }

  protected void fireModulesAdded() {
    for (Module module : myModuleModel.getModules()) {
      fireModuleAddedInWriteAction((ModuleEx)module);
    }
  }

  protected void fireModuleAddedInWriteAction(@NotNull ModuleEx module) {
    ApplicationManager.getApplication().runWriteAction(() -> {
      if (!module.isLoaded()) {
        module.moduleAdded();
        myProject.getMessageBus().syncPublisher(ProjectTopics.MODULES).moduleAdded(myProject, module);
      }
    });
  }

  public static void commitModelWithRunnable(@NotNull ModifiableModuleModel model, @NotNull Runnable runnable) {
    ((ModuleModelImpl)model).commitWithRunnable(runnable);
  }

  @NotNull
  protected abstract ModuleEx createModule(@NotNull String filePath);

  protected ModuleEx createNonPersistentModule(@NotNull String name) {
    throw new UnsupportedOperationException();
  }

  @NotNull
  protected abstract ModuleEx createAndLoadModule(@NotNull String filePath) throws IOException;

  final static class ModuleModelImpl implements ModifiableModuleModel {
    final Map<String, Module> myModules;
    private volatile Module[] myModulesCache;

    private final List<Module> myModulesToDispose = new ArrayList<>();
    private final Map<Module, String> myModuleToNewName = new HashMap<>();
    private final Map<String, Module> myNewNameToModule = new HashMap<>();
    private boolean myIsWritable;
    private Map<Module, String[]> myModuleGroupPath;

    private final ModuleManagerImpl myManager;

    private ModuleModelImpl(@NotNull ModuleManagerImpl manager) {
      myManager = manager;
      myIsWritable = false;
      myModules = Collections.synchronizedMap(new LinkedHashMap<>());
    }

    @SuppressWarnings("CopyConstructorMissesField")
    private ModuleModelImpl(@NotNull ModuleModelImpl that) {
      myManager = that.myManager;
      myModules = Collections.synchronizedMap(new LinkedHashMap<>(that.myModules));
      Map<Module, String[]> groupPath = that.myModuleGroupPath;
      if (groupPath != null) {
        myModuleGroupPath = new HashMap<>(that.myModuleGroupPath);
      }
      myIsWritable = true;
    }

    private void assertWritable() {
      LOG.assertTrue(myIsWritable, "Attempt to modify committed ModifiableModuleModel");
    }

    @Override
    public Module @NotNull [] getModules() {
      Module[] cache = myModulesCache;
      if (cache == null) {
        Collection<Module> modules = myModules.values();
        cache = modules.toArray(Module.EMPTY_ARRAY);
        myModulesCache = cache;
      }
      return cache;
    }

    private Module @NotNull [] getSortedModules() {
      Module[] allModules = getModules().clone();
      if (allModules.length > 1) {
        Arrays.sort(allModules, moduleDependencyComparator());
      }
      return allModules;
    }

    @Override
    public void renameModule(@NotNull Module module, @NotNull String newName) throws ModuleWithNameAlreadyExists {
      final Module oldModule = getModuleByNewName(newName);
      myNewNameToModule.remove(myModuleToNewName.get(module));
      if(module.getName().equals(newName)){ // if renaming to itself, forget it altogether
        myModuleToNewName.remove(module);
        myNewNameToModule.remove(newName);
      } else {
        myModuleToNewName.put(module, newName);
        myNewNameToModule.put(newName, module);
      }

      if (oldModule != null) {
        throw new ModuleWithNameAlreadyExists(ProjectModelBundle.message("module.already.exists.error", newName), newName);
      }
    }

    @Override
    public Module getModuleToBeRenamed(@NotNull String newName) {
      return myNewNameToModule.get(newName);
    }

    private Module getModuleByNewName(@NotNull String newName) {
      final Module moduleToBeRenamed = getModuleToBeRenamed(newName);
      if (moduleToBeRenamed != null) {
        return moduleToBeRenamed;
      }
      final Module moduleWithOldName = findModuleByName(newName);
      return myModuleToNewName.get(moduleWithOldName) == null ? moduleWithOldName : null;
    }

    @Override
    public String getNewName(@NotNull Module module) {
      return myModuleToNewName.get(module);
    }

    @NotNull
    @Override
    public String getActualName(@NotNull Module module) {
      return ObjectUtils.notNull(getNewName(module), module.getName());
    }

    @Override
    @NotNull
    public Module newModule(@NotNull String filePath, @NotNull String moduleTypeId) {
      return newModule(filePath, null, moduleTypeId, null);
    }

    @Override
    public Module newModule(@NotNull Path file, @NotNull String moduleTypeId) {
      return newModule(null, file, moduleTypeId, null);
    }

    @Override
    @NotNull
    public Module newNonPersistentModule(@NotNull String moduleName, @NotNull final String moduleTypeId) {
      assertWritable();

      ModuleEx module = myManager.createNonPersistentModule(moduleName);
      initModule(module, () -> {
        module.setModuleType(moduleTypeId);
      });
      addModule(module);
      return module;
    }

    @Override
    @NotNull
    public Module newModule(@NotNull String filePath, @NotNull String moduleTypeId, @Nullable Map<String, String> options) {
      return newModule(filePath, null, moduleTypeId, options);
    }

    private Module newModule(@Nullable String filePath, @Nullable Path file, @NotNull String moduleTypeId, @Nullable Map<String, String> options) {
      assertWritable();
      if (filePath == null) {
        filePath = Objects.requireNonNull(file).toAbsolutePath().normalize().toString().replace(File.separatorChar, '/');
      }
      else {
        filePath = FileUtil.toSystemIndependentName(resolveShortWindowsName(filePath));
      }

      ModuleEx module = getModuleByFilePath(filePath);
      if (module != null) {
        return module;
      }

      module = myManager.createModule(filePath);
      ModuleEx newModule = module;
      Path finalFile = file == null ? Paths.get(filePath) : file;
      initModule(module, () -> {
        ((ModuleStore)newModule.getService(IComponentStore.class)).setPath(finalFile, null, true);

        newModule.setModuleType(moduleTypeId);
        if (options != null) {
          for (Map.Entry<String, String> option : options.entrySet()) {
            //noinspection deprecation
            newModule.setOption(option.getKey(), option.getValue());
          }
        }
      });
      addModule(module);
      return module;
    }

    @NotNull
    private static String resolveShortWindowsName(@NotNull String filePath) {
      try {
        return FileUtil.resolveShortWindowsName(filePath);
      }
      catch (IOException ignored) {
        return filePath;
      }
    }

    private @Nullable ModuleEx getModuleByFilePath(@NotNull @SystemIndependent String filePath) {
      for (Module module : getModules()) {
        if (SystemInfo.isFileSystemCaseSensitive ? module.getModuleFilePath().equals(filePath) : module.getModuleFilePath().equalsIgnoreCase(filePath)) {
          return (ModuleEx)module;
        }
      }
      return null;
    }

    @Override
    public Module loadModule(@NotNull Path file) throws IOException {
      assertWritable();
      Path normalizedPath = file.toAbsolutePath().normalize();
      try {
        Module module = getModuleByFilePath(FileUtilRt.toSystemIndependentName(normalizedPath.toString()));
        if (module == null) {
          module = loadModuleInternal(normalizedPath, myManager);
          addModule(module);
        }
        return module;
      }
      catch (NoSuchFileException e) {
        throw e;
      }
      catch (IOException e) {
        throw new IOException(ProjectModelBundle.message("module.corrupted.file.error", normalizedPath.toString(), e.getMessage()), e);
      }
    }

    @Override
    public @NotNull Module loadModule(@NotNull @SystemIndependent String filePath) throws IOException {
      return loadModule(Paths.get(resolveShortWindowsName(filePath)));
    }

    private void addModule(@NotNull Module module) {
      myModulesCache = null;
      myModules.put(module.getName(), module);
    }

    @Override
    public void disposeModule(@NotNull Module module) {
      assertWritable();
      myModulesCache = null;
      if (myModules.remove(module.getName()) != null) {
        myModulesToDispose.add(module);
      }
      if (myModuleGroupPath != null){
        myModuleGroupPath.remove(module);
      }
    }

    @Override
    public Module findModuleByName(@NotNull String name) {
      Module module = myModules.get(name);
      if (module != null && !module.isDisposed()) {
        return module;
      }
      return null;
    }

    @NotNull
    private Comparator<Module> moduleDependencyComparator() {
      DFSTBuilder<Module> builder = new DFSTBuilder<>(moduleGraph(true));
      return builder.comparator();
    }

    @NotNull
    private Graph<Module> moduleGraph(final boolean includeTests) {
      return GraphGenerator.generate(CachingSemiGraph.cache(new InboundSemiGraph<>() {
        @NotNull
        @Override
        public Collection<Module> getNodes() {
          return myModules.values();
        }

        @NotNull
        @Override
        public Iterator<Module> getIn(Module m) {
          Module[] dependentModules = ModuleRootManager.getInstance(m).getDependencies(includeTests);
          return Arrays.asList(dependentModules).iterator();
        }
      }));
    }

    @Override
    public void commit() {
      ModifiableModelCommitter.multiCommit(Collections.emptyList(), this);
    }

    private void commitWithRunnable(@NotNull Runnable runnable) {
      myManager.commitModel(this, runnable);
      clearRenamingStuff();
    }

    private void clearRenamingStuff() {
      myModuleToNewName.clear();
      myNewNameToModule.clear();
    }

    @Override
    public void dispose() {
      assertWritable();
      ApplicationManager.getApplication().assertWriteAccessAllowed();
      final Set<Module> existingModules = ContainerUtil.set(myManager.myModuleModel.getModules());
      for (Module thisModule : getModules()) {
        if (!existingModules.contains(thisModule)) {
          Disposer.dispose(thisModule);
        }
      }
      for (Module moduleToDispose : myModulesToDispose) {
        if (!existingModules.contains(moduleToDispose)) {
          Disposer.dispose(moduleToDispose);
        }
      }
      clearRenamingStuff();
    }

    @Override
    public boolean isChanged() {
      if (!myIsWritable) {
        return false;
      }
      ModuleModelImpl model = myManager.myModuleModel;
      return !myModules.equals(model.myModules)
             || !Comparing.equal(model.myModuleGroupPath, myModuleGroupPath)
             || !myModuleToNewName.isEmpty();
    }

    private void disposeModel() {
      Module[] modules = getModules();
      myModulesCache = null;
      // clear module list before disposing to avoid getModules() returning already disposed modules
      myModules.clear();
      for (Module module : modules) {
        Disposer.dispose(module);
      }
      myModuleGroupPath = null;
    }

    @Override
    public String[] getModuleGroupPath(@NotNull Module module) {
      return myModuleGroupPath == null ? null : myModuleGroupPath.get(module);
    }

    @Override
    public boolean hasModuleGroups() {
      return myModuleGroupPath != null && !myModuleGroupPath.isEmpty();
    }

    @Override
    public void setModuleGroupPath(@NotNull Module module, String @Nullable("null means remove") [] groupPath) {
      if (myModuleGroupPath == null) {
        myModuleGroupPath = new HashMap<>();
      }
      if (groupPath == null) {
        myModuleGroupPath.remove(module);
      }
      else {
        myModuleGroupPath.put(module, groupPath);
      }
    }

    @NotNull
    @Override
    public Project getProject() {
      return myManager.myProject;
    }
  }

  private void commitModel(@NotNull ModuleModelImpl moduleModel, @NotNull Runnable runnable) {
    myModuleModel.myModulesCache = null;
    incModificationCount();
    ApplicationManager.getApplication().assertWriteAccessAllowed();
    final Collection<Module> oldModules = Arrays.asList(myModuleModel.getModules());
    final Collection<Module> newModules = Arrays.asList(moduleModel.getModules());

    final Collection<Module> addedModules;
    final Collection<Module> removedModules;
    if (oldModules.isEmpty()) {
      addedModules = newModules;
      removedModules = Collections.emptyList();
    }
    else {
      addedModules = new HashSet<>(newModules);
      addedModules.removeAll(oldModules);

      removedModules = new HashSet<>(oldModules);
      removedModules.removeAll(newModules);
    }

    myProjectRootManager.makeRootsChange(() -> {
      for (Module removedModule : removedModules) {
        fireBeforeModuleRemoved(removedModule);
        cleanCachedStuff();
      }

      if (!moduleModel.myModulesToDispose.isEmpty()) {
        List<Module> neverAddedModules = new ArrayList<>(moduleModel.myModulesToDispose);
        neverAddedModules.removeAll(oldModules);
        for (final Module neverAddedModule : neverAddedModules) {
          neverAddedModule.putUserData(DISPOSED_MODULE_NAME, neverAddedModule.getName());
          Disposer.dispose(neverAddedModule);
        }
      }

      runnable.run();

      final Map<Module, String> modulesToNewNamesMap = moduleModel.myModuleToNewName;
      final Set<Module> modulesToBeRenamed = modulesToNewNamesMap.keySet();
      modulesToBeRenamed.removeAll(moduleModel.myModulesToDispose);

      List<Module> modules = new ArrayList<>();
      Map<Module, String> oldNames = new HashMap<>();
      for (final Module module : modulesToBeRenamed) {
        oldNames.put(module, module.getName());
        moduleModel.myModules.remove(module.getName());
        modules.add(module);
        ((ModuleEx)module).rename(modulesToNewNamesMap.get(module), true);
        moduleModel.myModules.put(module.getName(), module);
        myUnloadedModules.remove(module.getName());
      }

      moduleModel.myIsWritable = false;
      myModuleModel = moduleModel;

      for (Module module : removedModules) {
        fireModuleRemoved(module);
        cleanCachedStuff();
        Disposer.dispose(module);
        cleanCachedStuff();
      }

      ModuleListener publisher = myProject.getMessageBus().syncPublisher(ProjectTopics.MODULES);
      for (Module addedModule : addedModules) {
        myUnloadedModules.remove(addedModule.getName());
        ((ModuleEx)addedModule).moduleAdded();
        cleanCachedStuff();
        publisher.moduleAdded(myProject, addedModule);
        cleanCachedStuff();
      }
      cleanCachedStuff();
      fireModulesRenamed(modules, oldNames);
      cleanCachedStuff();
      UnloadedModulesListStorage unloadedModulesListStorage = UnloadedModulesListStorage.getInstance(myProject);
      setUnloadedModuleNames(ContainerUtil.filter(unloadedModulesListStorage.getUnloadedModuleNames(), myUnloadedModules::containsKey));
    }, false, true);
  }

  public void fireModuleRenamedByVfsEvent(@NotNull Module module, @NotNull String oldName) {
    Module moduleInMap = myModuleModel.myModules.remove(oldName);
    LOG.assertTrue(moduleInMap == null || moduleInMap == module);
    myModuleModel.myModules.put(module.getName(), module);
    incModificationCount();

    myProjectRootManager.makeRootsChange(() -> {
      fireModulesRenamed(Collections.singletonList(module), Collections.singletonMap(module, oldName));
    }, false, true);
  }

  @Override
  public String[] getModuleGroupPath(@NotNull Module module) {
    return myModuleModel.getModuleGroupPath(module);
  }

  @Override
  public boolean hasModuleGroups() {
    return myModuleModel.hasModuleGroups();
  }

  @NotNull
  @Override
  public Collection<ModuleDescription> getAllModuleDescriptions() {
    Module[] modules = getModules();
    List<ModuleDescription> descriptions = new ArrayList<>(modules.length + myUnloadedModules.size());
    for (Module module : modules) {
      descriptions.add(new LoadedModuleDescriptionImpl(module));
    }
    descriptions.addAll(myUnloadedModules.values());
    return descriptions;
  }

  @NotNull
  @Override
  public Collection<UnloadedModuleDescription> getUnloadedModuleDescriptions() {
    return Collections.unmodifiableCollection(myUnloadedModules.values());
  }

  @Override
  public Collection<ModulePath> getFailedModulePaths() {
    return Collections.unmodifiableSet(myFailedModulePaths);
  }

  @Nullable
  @Override
  public UnloadedModuleDescription getUnloadedModuleDescription(@NotNull String moduleName) {
    return myUnloadedModules.get(moduleName);
  }

  @NotNull
  @Override
  public ModuleGrouper getModuleGrouper(@Nullable ModifiableModuleModel model) {
    return ModuleGroupersKt.createGrouper(myProject, model);
  }

  @Override
  public void setUnloadedModules(@NotNull List<String> unloadedModuleNames) {
    if (myUnloadedModules.keySet().equals(new HashSet<>(unloadedModuleNames))) {
      //optimization
      return;
    }

    setUnloadedModuleNames(unloadedModuleNames);

    final ModifiableModuleModel model = getModifiableModel();
    Map<String, UnloadedModuleDescriptionImpl> toLoad = new LinkedHashMap<>(myUnloadedModules);
    myUnloadedModules.clear();
    for (String name : unloadedModuleNames) {
      if (toLoad.containsKey(name)) {
        myUnloadedModules.put(name, toLoad.remove(name));
      }
      else {
        Module module = findModuleByName(name);
        if (module != null) {
          ModuleDescription description = new LoadedModuleDescriptionImpl(module);
          ModuleSaveItem saveItem = new ModuleSaveItem(module, this);
          ModulePath modulePath = new ModulePath(saveItem.getModuleFilePath(), saveItem.getGroupPathString());
          VirtualFilePointerManager pointerManager = VirtualFilePointerManager.getInstance();
          List<VirtualFilePointer> contentRoots = ContainerUtil.map(ModuleRootManager.getInstance(module).getContentRootUrls(), url -> pointerManager.create(url, this, null));
          UnloadedModuleDescriptionImpl unloadedModuleDescription = new UnloadedModuleDescriptionImpl(modulePath, description.getDependencyModuleNames(), contentRoots);
          // we need to save module configuration before unloading, otherwise its settings will be lost
          StateStorageManagerKt.saveComponentManager(module);
          model.disposeModule(module);
          myUnloadedModules.put(name, unloadedModuleDescription);
        }
      }
    }
    List<ModulePath> oldFailedPaths = new ArrayList<>(myFailedModulePaths);
    myModulePathsToLoad = toLoad.values().stream()
      .map(UnloadedModuleDescriptionImpl::getModulePath)
      .collect(Collectors.toCollection(LinkedHashSet::new));
    loadModules((ModuleModelImpl)model);
    ApplicationManager.getApplication().runWriteAction(model::commit);
    myFailedModulePaths.addAll(oldFailedPaths);
    myModulePathsToLoad.clear();
  }

  @Override
  public void removeUnloadedModules(@NotNull Collection<? extends UnloadedModuleDescription> unloadedModules) {
    ApplicationManager.getApplication().assertWriteAccessAllowed();
    for (UnloadedModuleDescription module : unloadedModules) {
      myUnloadedModules.remove(module.getName());
    }
    setUnloadedModuleNames(new ArrayList<>(myUnloadedModules.keySet()));
  }

  protected void setUnloadedModuleNames(@NotNull List<String> unloadedModuleNames) {
    UnloadedModulesListStorage.getInstance(myProject).setUnloadedModuleNames(unloadedModuleNames);
  }

  public void setModuleGroupPath(@NotNull Module module, String @Nullable [] groupPath) {
    myModuleModel.setModuleGroupPath(module, groupPath);
  }
}

